/*
 *  matqMath.cpp
 *  fp
 *
 *  Created by Robert Delaney on 5/28/13.
 *  Copyright 2013 Bob Delaney's Science Software. All rights reserved.
 *
 */


#include "fp.h"
#include "mb.h"
#include "bf.h"
#include "matq.h"
#include "matqMath.h"
#include "matqConv.h"

// swap row i with row j
static void matqSwapRows(matq& x, INT32 i, INT32 j)
{
	bf		temp;
	INT32	k;
	
	for(k=0;k<x.nc;k++)
	{
		temp = x.array[i][k];
		x.array[i][k] = x.array[j][k];
		x.array[j][k] = temp;
	}
	
}/* matqSwapRows */


// returns 1 if x=y else returns 0
INT32 matqCompare(const matq& x, const matq& y)
{
	INT32		i, j;
	
	if(x.nr!=y.nr || x.nc!=y.nc)
		return 0;
	
	for(i=0;i<x.nr;i++)
		for(j=0;j<x.nc;j++)
			if(x.array[i][j]!=y.array[i][j])
				return 0;
	
	return 1;
	
}/* matqCompare */


void add(matq& z, const matq& x, const matq& y)
{
	matq		zt;
	INT32		i, j;
	
	if(x.nr!=y.nr || x.nc!=y.nc)
		return;
	
	zt.nr = x.nr;
	zt.nc = x.nc;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
			zt.array[i][j] = x.array[i][j] + y.array[i][j];
	
	z = zt;
	
}/* add */


void sub(matq& z, const matq& x, const matq& y)
{
	matq		zt;
	INT32		i, j;
	
	if(x.nr!=y.nr || x.nc!=y.nc)
		return;
	
	zt.nr = x.nr;
	zt.nc = x.nc;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
			zt.array[i][j] = x.array[i][j] - y.array[i][j];
	
	z = zt;
	
}/* sub */


void mul(matq& z, const matq& x, const matq& y)
{
	matq		zt;
	INT32		i, j, k;
	
	if(x.nc!=y.nr)
		return;
	
	zt.nr = x.nr;
	zt.nc = y.nc;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
		{
			zt.array[i][j] = 0;
			for(k=0;k<x.nc;k++)
				zt.array[i][j] += x.array[i][k] * y.array[k][j];
		}
	
	z = zt;
	
}/* mul */
	

void mul(matq& z, const matq& x, const bf& y)
{
	matq		zt;
	INT32		i, j;
	
	zt.nr = x.nr;
	zt.nc = x.nc;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
			zt.array[i][j] = y * x.array[i][j];
	
	z = zt;
	
}/* mul */

void mul(matq& z, const matq& x, const mb& y)
{
	matq		zt;
	INT32		i, j;
	
	zt.nr = x.nr;
	zt.nc = x.nc;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
			zt.array[i][j] = y * x.array[i][j];
	
	z = zt;
	
}/* mul */


void mul(matq& z, const matq& x, double y)
{
	matq		zt;
	INT32		i, j;
	
	zt.nr = x.nr;
	zt.nc = x.nc;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
			zt.array[i][j] = y * x.array[i][j];
	
	z = zt;
	
}/* mul */

// z = x*y mod q only for integer matrices
static void mulMod(matq& z, const matq& x, const matq& y, const mb& q)
{
	matq		zt;
	INT32		i, j, k;
	
	if(x.nc!=y.nr)
		return;
	
	zt.nr = x.nr;
	zt.nc = y.nc;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
		{
			zt.array[i][j] = 0;
			for(k=0;k<x.nc;k++)
			{
				zt.array[i][j] += x.array[i][k] * y.array[k][j];
				zt.array[i][j].num = (zt.array[i][j].num % q);
			}
		}
	
	z = zt;
	
}/* mulMod */



// creates unit matrix of size n
void matqUnit(matq& z, INT32 n)
{
	INT32		i, j;
	
	z.nr = z.nc = n;
	z.array = (bf**)malloc(z.nr*sizeof(bf*));
	for(i=0;i<z.nr;i++)
		z.array[i] = (bf*)malloc(z.nc*sizeof(bf));
	
	for(i=0;i<z.nr;i++)
		for(j=0;j<z.nc;j++)
		{
			if(i==j)
				z.array[i][j] = 1;
			else
				z.array[i][j] = 0;
		}
	
}/* matqUnit */


matq matqUnit(INT32 n)
{
	matq		z;
	
	matqUnit(z, n);
	
	return z;
	
}/* matqUnit */


void transpose(matq& z, const matq& x)
{
	matq		zt;
	INT32		i, j;
	
	zt.nr = x.nc;
	zt.nc = x.nr;
	
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	for(i=0;i<zt.nr;i++)
		for(j=0;j<zt.nc;j++)
			zt.array[i][j] = x.array[j][i];
	
	z = zt;
	
}/* transpose */


matq transpose(const matq& x)
{
	matq		z;
	
	transpose(z, x);
	
	return z;
	
}/* transpose */

// x must be square
bool trace(bf& z, const matq& x)
{
	bf			zt;
	INT32		i;
	
	if(x.nr != x.nc)
	{
		z = 0;
		return false;
	}
	
	zt = 0;
	for(i=0;i<x.nr;i++)
		zt+=x.array[i][i];
	
	z = zt;
	
	return true;
	
}/* trace */


bf trace(const matq& x)
{
	bf		z;
	
	if(trace(z, x))
	{
		return z;
	}
	else
	{
		z = 0;
		return z;
	}

}/* trace */


// puts matrix y.array at bottom of x.array; the matrices must have the same number of columns
void matqRowAugment(matq& z, const matq& x, const matq& y)
{
	matq		zt;
	INT32		i, j;
	
	if(x.nc!=y.nc)
		return;
	
	zt.nr = x.nr + y.nr;
	zt.nc = x.nc;
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	// copy x.array into zt.array top
	for(i=0;i<x.nr;i++)
		for(j=0;j<x.nc;j++)
			zt.array[i][j] = x.array[i][j];
	
	// copy y.array to bottom of zt.array
	for(i=0;i<y.nr;i++)
		for(j=0;j<y.nc;j++)
			zt.array[i+x.nr][j] = y.array[i][j];
	
	z = zt;
	
}/* matqRowAugment */


matq matqRowAugment(const matq& x, const matq& y)
{
	matq	z;
	
	matqRowAugment(z, x, y);
	
	return z;
	
}/* matqRowAugment */


// puts matrix y.array to right of x.array; the matrices must have the same number of rows
void matqColAugment(matq& z, const matq& x, const matq& y)
{
	matq		zt;
	INT32		i, j;
	
	if(x.nr!=y.nr)
		return;
	
	zt.nr = x.nr;
	zt.nc = x.nc + y.nc;
	zt.array = (bf**)malloc(zt.nr*sizeof(bf*));
	for(i=0;i<zt.nr;i++)
		zt.array[i] = (bf*)malloc(zt.nc*sizeof(bf));
	
	// copy x.array into zt.array left
	for(i=0;i<x.nr;i++)
		for(j=0;j<x.nc;j++)
			zt.array[i][j] = x.array[i][j];
	
	// copy y.array to right of zt.array
	for(i=0;i<y.nr;i++)
		for(j=0;j<y.nc;j++)
			zt.array[i][j+x.nc] = y.array[i][j];
	
	z = zt;
	
}/* matqColAugment */

matq matqColAugment(const matq& x, const matq& y)
{
	matq	z;
	
	matqColAugment(z, x, y);
	
	return z;
	
}/* matqColAugment */

// by rows
void matqGaussElim(matq& z, INT32& swapParity, const matq& x)
{
	matq	xt;
	INT32	i, j, k, i_max, ii;
	bool	inSearch;
	
	swapParity = 1;
	
	xt = x;
	ii = 0; // i -> pivot row; ii -> pivot column
	for(i=0;i<xt.nr-1;i++)
	{
		inSearch = true;
		while(inSearch)
		{
			// find row whose iith column is not zero at ith row and below
			i_max = i;
			for(k=i+1;k<xt.nr;k++)
			{
				if(abs(xt.array[k][ii])>abs(xt.array[i_max][ii]))
					i_max = k;
			}
			
			if(!xt.array[i_max][ii].num.n)
			{
				// iith column must be skipped since it's zero from ith row down
				ii++; // move to next column to right
				if(ii>=xt.nc)
				{
					z = xt;
					return;
				}
			}
			else
				inSearch = false;
		}
	
		if(!(i==i_max))
		{
			matqSwapRows(xt, i, i_max);
			swapParity = -swapParity;
		}

		// Eliminate the iith element of the rows below the ith
		
		for(k=i+1;k<xt.nr;k++)
		{
			for(j=ii+1;j<xt.nc;j++)
				xt.array[k][j] = xt.array[k][j] - xt.array[k][ii] * xt.array[i][j] / xt.array[i][ii];
			xt.array[k][ii] = 0;
		}
		ii++;
		if(ii>=xt.nc)
		{
			z = xt;
			return;
		}
	}
	
	z = xt;
	
}/* matqGaussElim */


bool matqGaussElim(matq& z, INT32& swapParity, const char *xString)
{
	matq		xt;
	
	if(!matqConvFromStr(xt, xString))
		return false;
	
	matqGaussElim(z, swapParity, xt);
	
	return true;
	
}/* matqGaussElim */


// by rows mod q
void matqGaussElim(matq& z, const matq& x, double q)
{
	matq	xt;
	INT32	i, j, k, i_max, ii;
	bool	inSearch;
	
	xt = x;
	ii = 0; // i -> pivot row; ii -> pivot column
	for(i=0;i<xt.nr-1;i++)
	{
		inSearch = true;
		while(inSearch)
		{
			// find row whose iith column is not zero at ith row and below
			i_max = i;
			for(k=i+1;k<xt.nr;k++)
			{
				if(abs(xt.array[k][ii])>abs(xt.array[i_max][ii]))
					i_max = k;
			}
			
			if(!xt.array[i_max][ii].num.n)
			{
				// iith column must be skipped since it's zero from ith row down
				ii++; // move to next column to right
				if(ii>=xt.nc)
				{
					z = xt;
					return;
				}
			}
			else
				inSearch = false;
		}
		
		if(!(i==i_max))
		{
			matqSwapRows(xt, i, i_max);
		}
		
		// Eliminate the iith element of the rows below the ith
		
		for(k=i+1;k<xt.nr;k++)
		{
			for(j=ii+1;j<xt.nc;j++)
			{
				xt.array[k][j] = xt.array[i][ii] * xt.array[k][j] - xt.array[k][ii] * xt.array[i][j];
				xt.array[k][j].num = (xt.array[k][j].num % q);
			}
				
			xt.array[k][ii] = 0;
		}
		ii++;
		if(ii>=xt.nc)
		{
			z = xt;
			return;
		}
	}
	
	z = xt;
	
}/* matqGaussElim */


// x.array must be in echelon form
void matqReduce(matq& z, const matq& x)
{
	matq	xt;
	bf		temp;
	INT32	i, j, k, kMax;
	
	xt = x;
	
	kMax = min(xt.nr, xt.nc);
	
	for(k=0;k<kMax;k++)
	{
		temp = xt.array[k][k];
		if(temp.num.n)
			for(j=k;j<xt.nc;j++)
				xt.array[k][j] = xt.array[k][j] / temp;
	}
	
	// now ensure every diagonal 1 has zeroes above it
	for(k=1;k<kMax;k++)
	{
		temp = xt.array[k][k]; // it's 1 or 0
		
		if(temp.num.n)
		{
			for(i=0;i<k;i++)
			{
				temp = xt.array[i][k];
				for(j=k;j<xt.nc;j++)
					xt.array[i][j] = xt.array[i][j] - temp * xt.array[k][j];
			}
		}
	}
	
	z = xt;
	
}/* matqReduce */

void matqNullSpace(INT32& numNV, matq& nullVectors, matq& reduced, const matq& x)
{
	matq		xt, yt, zt;
	INT32		i, j, k;
	INT32		swapParity;
	
	xt = x;

	matqRowAugment(yt, xt, matqUnit(xt.nc));
	matqGaussElim(zt, swapParity, transpose(yt));
	matqReduce(reduced, zt);
	
	// look for zero rows in reduced for column index from 0 to xt.nr; null vectors are at the ends of these rows
	// create a Boolean array to hold true or false for null vector rows
	bool	*hasNV;
	bool	isNull;
	hasNV = (bool*)malloc(reduced.nr*sizeof(bool));
	for(i=0;i<reduced.nr;i++)
	{
		isNull = true;
		for(j=0;j<xt.nr;j++)
		{
			if(reduced.array[i][j].num.n)
			{
				isNull = false;
				break;
			}
		}
		if(isNull)
			hasNV[i] = true;
		else
			hasNV[i] = false;
	}
	
	// get number of null vectors
	numNV = 0;
	for(i=0;i<reduced.nr;i++)
	{
		if(hasNV[i])
			numNV++;
	}
	
	if(!numNV)
	{
		free(hasNV);
		nullVectors.nr = 1;
		nullVectors.nc = 1;
		nullVectors.array = (bf**)malloc(sizeof(bf*));
		nullVectors.array[0] = (bf*)malloc(sizeof(bf));
		nullVectors.array[0][0] = 0;
		
		return;
	}
	
	// now have null vectors which we want as columns of myNullVectors.array
	nullVectors.nr = xt.nc;
	nullVectors.nc = numNV;  // column size of xt
	nullVectors.array = (bf**)malloc(nullVectors.nr*sizeof(bf*));
	for(i=0;i<nullVectors.nr;i++)
		nullVectors.array[i] = (bf*)malloc(nullVectors.nc*sizeof(bf));
	
	// copy null vectors into columns of nullVectors.array
	k = 0;
	for(j=0;j<xt.nc;j++)
	{
		if(hasNV[j])
		{
			for(i=0;i<nullVectors.nr;i++)
				nullVectors.array[i][k] = reduced.array[j][xt.nr+i];
			k++;
		}
	}
	
	mb		temp;
	
	for(j=0;j<nullVectors.nc;j++)
	{
		temp = 1;
		for(i=0;i<nullVectors.nr;i++)
			temp = temp*nullVectors.array[i][j].den;
		for(i=0;i<nullVectors.nr;i++)
			nullVectors.array[i][j] = temp*nullVectors.array[i][j];
		if(nullVectors.nr==1)
			nullVectors.array[0][j] = 1;
		else
		{
			temp = gcd(nullVectors.array[0][j].num,nullVectors.array[1][j].num);
			for(i=2;i<nullVectors.nr;i++)
				temp = gcd(temp, nullVectors.array[i][j].num);
			for(i=0;i<nullVectors.nr;i++)
				nullVectors.array[i][j] = nullVectors.array[i][j] / temp;
		}

	}
	
	free(hasNV);

}/* matqNullSpace */


bool matqNullSpace(INT32& numNV, matq& nullVectors, matq& reduced, const char *xString)
{
	matq		x;
	
	if(!matqConvFromStr(x, xString))
		return false;
	
	matqNullSpace(numNV, nullVectors, reduced, x);
	
	return true;
	
}/* matqNullSpace */

// mod q
void matqNullSpace(INT32& numNV, matq& nullVectors, matq& reduced, const matq& x, double q)
{
	matq		xt, yt, zt;
	INT32		i, j, k;
	
	xt = x;
	
	matqRowAugment(yt, xt, matqUnit(xt.nc));
	matqGaussElim(zt, transpose(yt), q);
	matqReduce(reduced, zt);
	
	// look for zero rows in reduced for column index from 0 to xt.nr; null vectors are at the ends of these rows
	// create a Boolean array to hold true or false for null vector rows
	bool	*hasNV;
	bool	isNull;
	hasNV = (bool*)malloc(reduced.nr*sizeof(bool));
	for(i=0;i<reduced.nr;i++)
	{
		isNull = true;
		for(j=0;j<xt.nr;j++)
		{
			if(reduced.array[i][j].num.n)
			{
				isNull = false;
				break;
			}
		}
		if(isNull)
			hasNV[i] = true;
		else
			hasNV[i] = false;
	}
	
	// get number of null vectors
	numNV = 0;
	for(i=0;i<reduced.nr;i++)
	{
		if(hasNV[i])
			numNV++;
	}
	
	if(!numNV)
	{
		free(hasNV);
		nullVectors.nr = 1;
		nullVectors.nc = 1;
		nullVectors.array = (bf**)malloc(sizeof(bf*));
		nullVectors.array[0] = (bf*)malloc(sizeof(bf));
		nullVectors.array[0][0] = 0;
		
		return;
	}
	
	// now have null vectors which we want as columns of myNullVectors.array
	nullVectors.nr = xt.nc;
	nullVectors.nc = numNV;  // column size of xt
	nullVectors.array = (bf**)malloc(nullVectors.nr*sizeof(bf*));
	for(i=0;i<nullVectors.nr;i++)
		nullVectors.array[i] = (bf*)malloc(nullVectors.nc*sizeof(bf));
	
	// copy null vectors into columns of nullVectors.array
	k = 0;
	for(j=0;j<xt.nc;j++)
	{
		if(hasNV[j])
		{
			for(i=0;i<nullVectors.nr;i++)
				nullVectors.array[i][k] = reduced.array[j][xt.nr+i];
			k++;
		}
	}
	
	for(i=0;i<nullVectors.nr;i++)
		for(j=0;j<nullVectors.nc;j++)
			nullVectors.array[i][j].num = (nullVectors.array[i][j].num % q);
	
	mb		temp;
	
	for(j=0;j<nullVectors.nc;j++)
	{
		temp = 1;
		for(i=0;i<nullVectors.nr;i++)
			temp = temp*nullVectors.array[i][j].den;
		for(i=0;i<nullVectors.nr;i++)
			nullVectors.array[i][j] = temp*nullVectors.array[i][j];
		if(nullVectors.nr==1)
			nullVectors.array[0][j] = 1;
		else
		{
			temp = gcd(nullVectors.array[0][j].num,nullVectors.array[1][j].num);
			for(i=2;i<nullVectors.nr;i++)
				temp = gcd(temp, nullVectors.array[i][j].num);
			if(temp>1)
				for(i=0;i<nullVectors.nr;i++)
					nullVectors.array[i][j] = nullVectors.array[i][j] / temp;
		}
	}
	
	free(hasNV);
	
}/* matqNullSpace */


bool matqInvert2(matq& z, bf& det, const matq& x)
{
	matq		xt;
	INT32		i, j;
	INT32		swapParity;
	
	if(x.nr!=x.nc)
		return false;
	
	if(x.nr==1)
	{
		z = x;
		if(!z.array[0][0].num.n)
			return false;
		
		z.array[0][0] = 1 / z.array[0][0];
	}
	
	xt = matqColAugment(x, matqUnit(x.nr));
	
	matqGaussElim(xt, swapParity, xt);
	det = swapParity;
	for(i=0;i<x.nr;i++)
		det = det * xt.array[i][i];
	
	matqReduce(xt, xt);
	
	z = x; // creates correct size matrux
	
	for(i=0;i<z.nr;i++)
		for(j=0;j<z.nc;j++)
			z.array[i][j] = xt.array[i][j+z.nc];
	
	// need to verify that the left square matrix is a unit matrix
	for(i=0;i<x.nr;i++)
		for(j=0;j<x.nr;j++)
		{
			if(i==j)
			{
				if(!(xt.array[i][j]==1))
					return false;
			}
			else
			{
				if(xt.array[i][j].num.n)
					return false;
			}
		}
	
	return true;
	
}/* matqInvert2 */


bool matqInvert(matq& z, bf& det, const matq& x)
{
	matq	zt;
	long	n, i, j, k, L, *ik, *jk;
	bf		amax, save, zero, one;
	
	zt = x;
	
	if(x.nr!=x.nc)
	{
		z = zt;
		return false;
	}
	
	n = x.nr;
	
	zero = 0;
	one = 1;
	
	ik = (long *)malloc(n*sizeof(long));
	jk = (long *)malloc(n*sizeof(long));
	
	det = one;
	
	for(k=0; k<n; k++)
	{
		/* find largest element in abs value in rest of matrix */
		amax = zero;
		
		for(i=k; i<n; i++)
		{
			for(j=k; j<n; j++)
			{
				if(abs(zt.array[i][j]) > abs(amax))
				{
					amax = zt.array[i][j];
					ik[k] = i;
					jk[k] = j;
				}
			}
		}
		
		if(abs(amax) == zero)
		{
			det = 0;
			return false;
		}
		
		/* interchange rows and columns to put amax in array(k,k) */
		i = ik[k];
		if(i>k)
		{
			for(j=0; j<n; j++)
			{
				save = zt.array[k][j];
				zt.array[k][j] = zt.array[i][j];
				zt.array[i][j] = -save;
			}
		}
		
		j = jk[k];
		if(j>k)
		{
			for(i=0; i<n; i++)
			{
				save = zt.array[i][k];
				zt.array[i][k] = zt.array[i][j];
				zt.array[i][j] = -save;
			}
		}
		
		/* accumulate elements of inverse matrix */
		for(i=0; i<n; i++)
		{
			if(i!=k)
				zt.array[i][k] = -zt.array[i][k]/amax;
		}
		
		for(i=0; i<n; i++)
		{
			for(j=0; j<n; j++)
			{
				if(i!=k && j!=k)
					zt.array[i][j] = zt.array[i][j] + zt.array[i][k]*zt.array[k][j];
			}
		}
		
		for(j=0; j<n; j++)
		{
			if(j!=k)
				zt.array[k][j] = zt.array[k][j]/amax;
		}
		
		zt.array[k][k] = one/amax;
		det = det*amax;
	}
	
	/* restore ordering of the matrix */
	for(L=0; L<n; L++)
	{
		k = n - L - 1;
		j = ik[k];
		
		if(j>k)
		{
			for(i=0; i<n; i++)
			{
				save = zt.array[i][k];
				zt.array[i][k] = -zt.array[i][j];
				zt.array[i][j] = save;
			}
		}
		
		i = jk[k];
		
		if(i>k)
		{
			for(j=0; j<n; j++)
			{
				save = zt.array[k][j];
				zt.array[k][j] = -zt.array[i][j];
				zt.array[i][j] = save;
			}
		}
	}
	
	free(ik);
	free(jk);
	
	z = zt;
	
	return true;
	
}/* matqInvert */


matq matqInvert(bf& det, const matq& x)
{
	matq		z;
	
	matqInvert(z, det, x);
	
	return z;
	
}/* matqInvert */


// meant for integer matrices
void matqMod(matq& z, const matq& x, const mb& q)
{
	INT32		i, j;
	
	z = x;
	
	for(i=0;i<x.nr;i++)
		for(j=0;j<x.nc;j++)
			z.array[i][j].num = z.array[i][j].num % q;
	
}/* matqMod */

// meant for integer matrices
void matqMod(matq& z, const matq& x, double q)
{
	INT32		i, j;
	
	z = x;
	
	for(i=0;i<x.nr;i++)
		for(j=0;j<x.nc;j++)
			z.array[i][j].num = z.array[i][j].num % q;

}/* matqMod */


// z = x^y ; x must be a square matrix and y>=0
bool power(matq& z, const matq& x, INT32 y)
{
	INT32				i, n, sy;
	static UINT32		mask; // 0x80000000
	//static mb			one;
	matq				zt;
	static bool			initGood=false;
	
	if(!initGood)
	{
		//init(one, 1);
		mask = 1;
		mask = (mask<<31); // 0x80000000
		initGood = true;
	}
	
	if(x.nr!=x.nc || y<0)
	{
		z = 0;
		return false;
	}
		
	if(matqCompare(x, matqUnit(x.nr)))
	{
		z = matqUnit(x.nr);
		return true;
	}
	
	if(y==0)
	{
		z = matqUnit(x.nr);
		return true;
	}
	
	zt = matqUnit(x.nr);
	sy = y;
	n = NumBits(sy);
	
	sy = (sy<<(32-n));  // go to leading 1 bit
	
	for(i=0;i<n;i++)
	{
		mul(zt, zt, zt); // zt^2
		if(sy & mask)
			mul(zt, zt, x);
		sy = (sy<<1);
	}
	
	z = zt;
	
	return true;
	
}/* power */


// z = x^y mod q ; x must be a square matrix and y>=0
bool powerMod(matq& z, const matq& x, const mb& y, const mb& q)
{
	INT32				i, n;
	mb					mask;
	matq				zt;
	
	if(x.nr!=x.nc || y.n<0)
	{
		z = 0;
		return false;
	}
	
	if(matqCompare(x, matqUnit(x.nr)))
	{
		z = matqUnit(x.nr);
		return true;
	}
	
	if(y.n==0)
	{
		z = matqUnit(x.nr);
		return true;
	}
	
	zt = matqUnit(x.nr);
	n = NumBits(y);
	
	// need to make mask have n bits with leading bit one and all others zero
	mask = 1;
	mask = (mask<<(n-1));
	
	for(i=0;i<n;i++)
	{
		mul(zt, zt, zt); // zt^2
		zt = zt % q;
		if((y & mask).n)
		{
			mul(zt, zt, x);
			zt = zt % q;
		}
		mask = (mask>>1);
	}
	
	z = zt;
	
	return true;
	
}/* powerMod */

